import logging
import ctypes as ct
from far2l.plugin import PluginBase

import debugpy


log = logging.getLogger(__name__)


class Plugin(PluginBase):
    label = "Python INFO"
    openFrom = ["PLUGINSMENU", "EDITOR"]

    # @staticmethod
    # def HandleCommandLine(line):
    #    return line in ('edit',)

    def getEditorInfo(self):
        ei = self.ffi.new("struct EditorInfo *")
        self.info.EditorControl(self.ffic.ECTL_GETINFO, ei)
        result = """\
EditorID: {}
WindowSizeX: {}
WindowSizeY: {}
TotalLines: {}
CurLine: {}
CurPos: {}
CurTabPos: {}
TopScreenLine: {}
LeftPos: {}
Overtype: {}
BlockType: {}
BlockStartLine: {}
Options: {}
TabSize: {}
BookMarkCount: {}
CurState: {}
CodePage: {}
Reserved[5]: {}
""".format(
            ei.EditorID,
            ei.WindowSizeX,
            ei.WindowSizeY,
            ei.TotalLines,
            ei.CurLine,
            ei.CurPos,
            ei.CurTabPos,
            ei.TopScreenLine,
            ei.LeftPos,
            ei.Overtype,
            ei.BlockType,
            ei.BlockStartLine,
            ei.Options,
            ei.TabSize,
            ei.BookMarkCount,
            ei.CurState,
            ei.CodePage,
            ei.Reserved,
        )
        return result

    def Info(self):
        # debugpy.breakpoint()
        txt = self.getEditorInfo()
        open("/tmp/uinfo", "wt").write(txt)
        self.info.Editor(
            "/tmp/uinfo",
            "uinfo",
            #            0, 0, -1, -1,
            0,
            0,
            25,
            25,
            self.ffic.EF_NONMODAL,
            0,
            0,
            0xFFFFFFFF,  # =-1=self.ffic.CP_AUTODETECT
        )
        # self.info.EditorControl(self.ffic.ECTL_INSERTTEXT, txt)

    def OpenPlugin(self, OpenFrom):
        # debugpy.breakpoint()
        if OpenFrom == 5:
            # EDITOR
            self.Info()
        return -1



__all__ = ["Window", "HSizer", "VSizer"]


class Orientation:
    horizontal = 0
    vertical = 1


class Window:
    def __init__(self, pos=(0, 0), size=(0, 0)):
        self.pos = pos
        self.size = size

    def move(self, l, t, w, h):
        self.pos = (l, t)
        self.size = (w, h)

    def get_best_size(self):
        raise NotImplementedError("Window.get_best_size")


class Box:
    def __init__(self, window, border):
        self.window = window
        self.border = border
        assert len(border) == 4

    def size(self, l, t, r, b):
        w = r - l
        h = b - t
        self.window.move(l, t, w, h)

    def get_best_size(self):
        x, y = self.window.get_best_size()
        return (
            x + self.border[0] + self.border[2],
            y + self.border[1] + self.border[3],
        )


class Sizer:
    orientation = None

    def __init__(self, border=(0, 0, 0, 0)):
        assert self.orientation in [Orientation.horizontal, Orientation.vertical]
        self.boxes = []
        self.border = border

    def add(self, w, border=(0, 0, 0, 0)):
        self.boxes.append(Box(w, border))

    def move(self, x, y, w, h):
        self.size(x, y, w, h)

    def size(self, l, t, r, b):
        l += self.border[0]
        t += self.border[1]
        r -= self.border[2]
        b -= self.border[3]
        hoffset = l
        voffset = t
        for box in self.boxes:
            cx, cy = box.get_best_size()
            if self.orientation == Orientation.horizontal:
                box.size(hoffset, voffset, hoffset + cx, voffset + cy)
                hoffset += cx
            elif self.orientation == Orientation.vertical:
                box.size(hoffset, voffset, hoffset + cx, voffset + cy)
                voffset += cy

    def get_best_size(self):
        b_x = 0
        b_y = 0
        for box in self.boxes:
            cx, cy = box.get_best_size()
            if self.orientation == Orientation.horizontal:
                b_x += cx
                b_y = max(cy, b_y)
            elif self.orientation == Orientation.vertical:
                b_x = max(cx, b_x)
                b_y += cy
        return b_x, b_y


class HSizer(Sizer):
    orientation = Orientation.horizontal


class VSizer(Sizer):
    orientation = Orientation.vertical



#!/usr/bin/env vpython3
import os
import stat
import time
from datetime import datetime
import io
import subprocess
import re
import logging

log = logging.getLogger(__name__)


class Entry(object):
    def __init__(
        self,
        dirname=None,
        perms=None,
        links=0,
        uid=0,
        gid=0,
        size=None,
        date=None,
        name=None,
        date_re=None,
    ):
        self.mode = 0
        self.perms = perms
        self.links = int(links)
        self.uid = int(uid)
        self.gid = int(gid)
        self.size = int(size)
        self.date = datetime.strptime(date, date_re)
        self.name = name

        if (self.perms[0] if self.perms else None) == "l" and " -> " in self.name:
            try:
                name, target = self.name.split(" -> ")
            except ValueError:
                return
            self.name = name

        self.perms2mode()

    def perms2mode(self):
        perms = self.perms[1:]
        lperms = len(perms)
        mode = 0
        # Check if 'perms' has the right format
        if not [x for x in perms if x not in "-rwxXsStT"] and lperms == 9:
            pos = lperms - 1
            for c in perms:
                if c in "sStT":
                    # Special modes
                    mode += (1 << pos // 3) << 9

                mode += 1 << pos if c in "rwxst" else 0
                pos -= 1
        self.mode = (
            mode
            | {
                "-": 0,
                "b": stat.S_IFBLK,
                "c": stat.S_IFCHR,
                "d": stat.S_IFDIR,
                "l": stat.S_IFLNK,
                "p": 0,  # stat.S_IF???,
                "s": 0,  # stat.S_IF???,
            }[self.perms[0]]
        )

    def __str__(self):
        if not self.name:
            return ""

        template = (
            "{mode:5o} {perms} {links:>4} {uid:<8} {gid:<8} {size:>8} {date} {name}"
        )
        return template.format(**self.__dict__)


class Docker(object):
    def __init__(self):
        pass

    def run(self, *args, timeout=2):
        cmd = ["/usr/bin/docker"]
        cmd.extend(args)
        # log.debug('docker.run: {}'.format(cmd))
        res = subprocess.run(cmd, capture_output=True, timeout=timeout)
        log.debug("rc={} stderr={}".format(res.returncode, res.stderr.decode()))
        res.check_returncode()
        assert res.stderr == b""
        fp = io.BytesIO(res.stdout)
        lines = fp.readlines()
        return lines

    def list(self):
        lines = self.run("container", "ps", "-a")
        devices = []
        if len(lines) > 0 and lines[0].split()[0] == b"CONTAINER":
            line = lines[0].decode()
            cid = 0
            cimage = line.find("IMAGE", 0)
            ccommand = line.find("COMMAND", cimage)
            ccreated = line.find("CREATED", ccommand)
            cstatus = line.find("STATUS", ccreated)
            cports = line.find("PORTS", cstatus)
            cnames = line.find("NAMES", cports)
            for line in lines[1:]:
                line = line.decode()
                info = (line[cid:cimage].strip(), line[cnames:].strip())
                devices.append(info)
        return devices

    ls = "/bin/ls -anL --full-time {}"
    date_re = "%Y-%m-%d %H:%M"
    file_re = (
        "^"
        "(?P<perms>[-bcdlps][-rwxsStT]{9})\s+"
        "(?P<links>\d+)\s"
        "(?P<uid>\d+)\s+"
        "(?P<gid>\d+)\s+"
        "(?P<size>\d+)\s"
        "(?P<date>\d{4}-\d{2}-\d{2}\s\d{2}:\d{2})\s"
        "(?P<name>.*)"
    )

    def ls(self, deviceid, top):
        file_re = re.compile(self.file_re)
        lines = self.run(
            "exec", deviceid, "/bin/ls", "-anL", "--time-style=long-iso", top
        )
        result = []
        for line in lines:
            line = line.strip()
            if not line:
                continue
            line = line.decode()
            rm = file_re.match(line)
            if not rm:
                if line.split()[0] != "total":
                    log.debug("ignored entry in ({}): {}".format(top, line))
                continue
            entry = Entry(**rm.groupdict(), date_re=self.date_re)
            result.append(entry)
        return result

    def pull(self, deviceid, sqname, dqname):
        lines = self.run("cp", "{}:{}".format(deviceid, sqname), dqname)
        for line in lines:
            log.debug("pull:".format(line))

    def push(self, deviceid, sqname, dqname):
        lines = self.run("cp", sqname, "{}:{}".format(deviceid, dqname))
        for line in lines:
            log.debug("push:".format(line))


if __name__ == "__main__":
    cls = Docker()
    info = cls.list()
    print(info)
    result = cls.ls(info[0][0], "/")
    for e in result:
        print(e)



#!/usr/bin/env vpython3
import sys
import os
import re
import pcpp
import io
import cffi

source = sys.argv[1]
target = os.path.join(sys.argv[2], "far2lcffi.py")

cpp = pcpp.Preprocessor()
cpp.add_path(source)
cpp.add_path(os.path.join(source, "far2l", "far2sdk"))
cpp.add_path(sys.argv[2])  # to find consts.h
cpp.define("FAR_USE_INTERNALS 1")
# cpp.define('WINPORT_DIRECT')
cpp.define("WINPORT_REGISTRY")
cpp.define("UNICODE")
cpp.define("PROCPLUGINMACROFUNC")
cpp.define("__GNUC__")
cpp.define("_FAR_NO_NAMELESS_UNIONS")
cpp.define("_FILE_OFFSET_BITS 64")
cpp.define("FAR_PYTHON_GEN 1")
data = """\
#define uid_t uint32_t
#define gid_t uint32_t
#include "consts.h"
#include "farplug-wide.h"
#include "farcolor.h"
#include "farkeys.h"
"""

cpp.parse(data)
fp = io.StringIO()
cpp.write(fp)

data = fp.getvalue()
for s in re.findall("'.+'", data):
    t = str(ord(s[-2].encode("ascii")))
    data = data.replace(s, t)
data = data.replace("#line", "//#line")
data = data.replace("#pragma", "//#pragma")

with open(target, 'wt') as fp:
    fp.write('''
''')
    fp.write('''\
data = """\\
''')
    fp.write(data)
    fp.write('''\
"""
''')
    fp.write('''\
import cffi
ffi = cffi.FFI()
ffi.cdef(data, packed=True)
del data
''')



import os
import stat
import time
import logging

from far2l.plugin import PluginVFS
from far2l.fardialogbuilder import (
    Spacer,
    TEXT,
    EDIT,
    PASSWORD,
    MASKED,
    MEMOEDIT,
    BUTTON,
    CHECKBOX,
    RADIOBUTTON,
    COMBOBOX,
    LISTBOX,
    USERCONTROL,
    HLine,
    HSizer,
    VSizer,
    DialogBuilder,
)

from udockerio import Docker

log = logging.getLogger(__name__)

TICKS_PER_SECOND = 10000000
EPOCH_DIFFERENCE = 11644473600

FILE_ATTRIBUTE_READONLY = 0x00000001
FILE_ATTRIBUTE_HIDDEN = 0x00000002
FILE_ATTRIBUTE_SYSTEM = 0x00000004
FILE_ATTRIBUTE_DIRECTORY = 0x00000010
FILE_ATTRIBUTE_ARCHIVE = 0x00000020
FILE_ATTRIBUTE_DEVICE = 0x00000040
FILE_ATTRIBUTE_NORMAL = 0x00000080
FILE_ATTRIBUTE_TEMPORARY = 0x00000100
FILE_ATTRIBUTE_SPARSE_FILE = 0x00000200
FILE_ATTRIBUTE_REPARSE_POINT = 0x00000400
FILE_ATTRIBUTE_COMPRESSED = 0x00000800
FILE_ATTRIBUTE_OFFLINE = 0x00001000
FILE_ATTRIBUTE_NOT_CONTENT_INDEXED = 0x00002000
FILE_ATTRIBUTE_ENCRYPTED = 0x00004000
FILE_ATTRIBUTE_INTEGRITY_STREAM = 0x00008000
FILE_ATTRIBUTE_VIRTUAL = 0x00010000
FILE_ATTRIBUTE_NO_SCRUB_DATA = 0x00020000
FILE_ATTRIBUTE_BROKEN = 0x00200000
FILE_ATTRIBUTE_EXECUTABLE = 0x00400000


class Plugin(PluginVFS):
    label = "Python Docker"
    openFrom = ["PLUGINSMENU", "DISKMENU"]

    def OpenPlugin(self, OpenFrom):
        self.names = []
        self.Items = []
        self.clt = Docker()
        self.device = None
        self.devicepath = "/"
        return True

    def devLoadDevices(self):
        for d in self.clt.list():
            self.addName(d[1], FILE_ATTRIBUTE_DIRECTORY, 0)

    def devSelectDevice(self, name):
        self.device = None
        for d in self.clt.list():
            if d[1] == name:
                self.device = d[0]
        self.devicepath = "/"

    def loadDirectory(self):
        result = self.clt.ls(self.device, self.devicepath)
        self.addResult(result)

    def devGetFile(self, sqname, dqname):
        self.clt.pull(self.device, sqname, dqname)

    def devPutFile(self, sqname, dqname):
        self.clt.push(self.device, sqname, dqname)

    def devDelete(self, dqname):
        pass

    def devMakeDirectory(self, dqname):
        pass

    def setName(self, i, name, attr, size):
        self.names[i] = self.s2f(name)
        item = self.Items[i].FindData
        item.dwFileAttributes = attr
        item.nFileSize = size
        item.lpwszFileName = self.names[i]
        return item

    def addName(self, name, attr, size):
        # log.debug('addName({})'.format(name))
        # increase C array
        items = self.ffi.new("struct PluginPanelItem []", len(self.Items) + 1)
        for i in range(len(self.Items)):
            items[i] = self.Items[i]
        self.Items = items
        self.names.append(None)
        self.setName(len(self.Items) - 1, name, attr, size)

    def addResult(self, result):
        n = 0
        for rec in result:
            if rec.name in (".", ".."):
                continue
            n += 1
        self.Items = self.ffi.new("struct PluginPanelItem []", n)
        self.names = [None] * n
        i = 0
        for rec in result:
            if rec.name in (".", ".."):
                continue
            attr = 0
            attr |= FILE_ATTRIBUTE_DIRECTORY if rec.mode & stat.S_IFDIR else 0
            attr |= FILE_ATTRIBUTE_DEVICE if rec.mode & stat.S_IFCHR else 0
            attr |= FILE_ATTRIBUTE_ARCHIVE if rec.mode & stat.S_IFREG else 0
            # log.debug('{} mode={:5o} attr={} perms={}'.format(rec.name, rec.mode, attr, rec.perms))
            # datetime.datetime.fromtimestamp(rec.time).strftime('%Y-%m-%d %H:%M:%S')
            item = self.setName(i, rec.name, attr, rec.size)
            item.dwUnixMode = rec.mode

            t = (
                int(time.mktime(rec.date.timetuple())) + EPOCH_DIFFERENCE
            ) * TICKS_PER_SECOND
            item.ftLastWriteTime.dwHighDateTime = t >> 32
            item.ftLastWriteTime.dwLowDateTime = t & 0xFFFFFFFF
            i += 1

    def deleteNames(self, names):
        found = []
        for i in range(len(self.names)):
            if self.f2s(self.names[i]) in names:
                found.append(i)
        if len(found) == 0:
            return False
        # new array
        n = len(self.Items) - len(found)
        if n > 0:
            items = self.ffi.new("struct PluginPanelItem []", n)
            j = 0
            for i in range(len(self.Items)):
                if i not in found:
                    items[j] = self.Items[i]
                    j += 1
            # delete
            for i in sorted(found, reverse=True):
                del self.names[i]
        else:
            self.names = []
            items = []
        self.Items = items
        self.info.Control(self.hplugin, self.ffic.FCTL_UPDATEPANEL, 0, 0)
        self.info.Control(self.hplugin, self.ffic.FCTL_REDRAWPANEL, 0, 0)
        return True

    def Message(self, lines):
        _MsgItems = [
            self.s2f("Docker"),
            self.s2f(""),
        ]
        for line in lines:
            _MsgItems.append(self.s2f(line))
        _MsgItems.extend(
            [
                self.s2f(""),
                self.s2f("\x01"),
                self.s2f("&Ok"),
            ]
        )
        # log.debug('_msgItems: %s', _MsgItems)
        MsgItems = self.ffi.new("wchar_t *[]", _MsgItems)
        self.info.Message(
            self.info.ModuleNumber,  # GUID
            self.ffic.FMSG_WARNING | self.ffic.FMSG_LEFTALIGN,  # Flags
            self.s2f("Contents"),  # HelpTopic
            MsgItems,  # Items
            len(MsgItems),  # ItemsNumber
            1,  # ButtonsNumber
        )

    def ConfirmDialog(self, title, lines):
        _MsgItems = [
            self.s2f(title),
            self.s2f(""),
        ]
        for line in lines:
            _MsgItems.append(self.s2f(line))
        _MsgItems.extend(
            [
                self.s2f(""),
                self.s2f("\x01"),
                self.s2f("&Cancel"),
                self.s2f("&Ok"),
            ]
        )
        # log.debug('_msgItems: %s', _MsgItems)
        MsgItems = self.ffi.new("wchar_t *[]", _MsgItems)
        rc = self.info.Message(
            self.info.ModuleNumber,  # GUID
            self.ffic.FMSG_WARNING | self.ffic.FMSG_LEFTALIGN,  # Flags
            self.s2f("Contents"),  # HelpTopic
            MsgItems,  # Items
            len(MsgItems),  # ItemsNumber
            2,  # ButtonsNumber
        )
        # 0 = Cancel, 1 = OK
        # log.debug('confirm={}'.format(rc))
        return rc

    def GetOpenPluginInfo(self, OpenInfo):
        if self.device is not None:
            title = "docker://{}{}".format(self.device, self.devicepath)
        else:
            title = self.label
        #
        # WARNING WARNING - dangerous pointers and lifetime of a variable in python
        # anything passed as a pointer must be kept in local python variables
        # otherwise gc/cffi will delete them and miracles will happen
        #
        self._curdir = self.s2f(self.devicepath)
        self._title = self.s2f(title)
        self._label = self.s2f(self.label)

        Info = self.ffi.cast("struct OpenPluginInfo *", OpenInfo)
        Info.Flags = (
            self.ffic.OPIF_USEFILTER
            | self.ffic.OPIF_USESORTGROUPS
            | self.ffic.OPIF_USEHIGHLIGHTING
            | self.ffic.OPIF_ADDDOTS
            | self.ffic.OPIF_SHOWNAMESONLY
        )
        Info.HostFile = self.ffi.NULL
        Info.CurDir = self._curdir
        Info.Format = self._label
        Info.PanelTitle = self._title
        # const struct InfoPanelLine *InfoLines;
        # int                   InfoLinesNumber;
        # const wchar_t * const   *DescrFiles;
        # int                   DescrFilesNumber;
        # const struct PanelMode *PanelModesArray;
        Info.PanelModesNumber = 0
        Info.StartPanelMode = 0
        # Info.StartSortMode = self.ffic.SM_NAME
        # Info.StartSortOrder = 0
        # const struct KeyBarTitles *KeyBar;
        # Info.ShortcutData = self.s2f('py:adb cd '+title)

    def GetFindData(self, PanelItem, ItemsNumber, OpMode):
        # super().GetFindData(PanelItem, ItemsNumber, OpMode)
        if OpMode & self.ffic.OPM_FIND:
            # find in root of device causes core dump on linux
            return False
        if self.device is None:
            try:
                self.devLoadDevices()
            except Exception as ex:
                log.exception("unknown exception:")
                msg = str(ex).split("\n")
                msg.insert(0, "Unknown exception.")
                self.Message(msg)
                return False
        else:
            try:
                self.loadDirectory()
            except Exception as ex:
                log.exception("unknown exception:")
                msg = str(ex).split("\n")
                msg.insert(0, "Unknown exception.")
                self.Message(msg)
                return False
        PanelItem = self.ffi.cast("struct PluginPanelItem **", PanelItem)
        ItemsNumber = self.ffi.cast("int *", ItemsNumber)
        n = len(self.Items)
        p = self.ffi.NULL if n == 0 else self.Items
        PanelItem[0] = p
        ItemsNumber[0] = n
        return True

    def FreeFindData(self, PanelItem, ItemsNumber):
        # log.debug("FreeFindData({0}, {1}, n.names={2}, n.Items={3})".format(PanelItem, ItemsNumber, len(self.names), len(self.Items)))
        self.names = []
        self.Items = []

    def SetDirectory(self, Dir, OpMode):
        # super().SetDirectory(Dir, OpMode)
        # if OpMode & self.ffic.OPM_FIND:
        #    return False
        name = self.f2s(Dir)
        # log.debug('goto.0: devicepath={} name={}'.format(self.devicepath, name))
        if name == "":
            self.info.Control(self.hplugin, self.ffic.FCTL_CLOSEPLUGIN, 0, 0)
            return True
        # log.debug('goto.1: devicepath={} name={}'.format(self.devicepath, name))
        if name == "..":
            if self.device is None:
                self.info.Control(self.hplugin, self.ffic.FCTL_CLOSEPLUGIN, 0, Dir)
                return True
            if self.devicepath == "/":
                self.device = None
            else:
                self.devicepath = os.path.normpath(os.path.join(self.devicepath, name))
            # log.debug('goto.2: devicepath={}'.format(self.devicepath))
        elif self.device is None:
            try:
                self.devSelectDevice(name)
            except Exception as ex:
                self.device = None
                log.exception("unknown exception:")
                msg = str(ex).split("\n")
                msg.insert(0, "Unknown exception.")
                self.Message(msg)
                return False
        else:
            self.devicepath = os.path.join(self.devicepath, name)
        return True

    def Compare(self, PanelItem1, PanelItem2, Mode):
        def cmp(a, b):
            if a < b:
                return -1
            elif a == b:
                return 0
            return 1

        p1 = self.ffi.cast("struct PluginPanelItem *", PanelItem1)
        p2 = self.ffi.cast("struct PluginPanelItem *", PanelItem2)
        n1 = self.f2s(p1.FindData.lpwszFileName)
        n2 = self.f2s(p2.FindData.lpwszFileName)
        rc = cmp(n1, n2)
        # log.debug('Compare: cmp({}, {})={}, mode={}'.format(n1, n2, rc, Mode))
        return rc

    def GetVirtualFindData(self, PanelItem, ItemsNumber, Path):
        return False

    def FreeVirtualFindData(self, PanelItem, ItemsNumber):
        log.debug("FreeVirtualFindData: {}, {}".format(PanelItem, ItemsNumber))

    def ProcessEvent(self, Event, Param):
        # if Event == self.ffic.FE_IDLE:
        pass

    def GetFiles(self, PanelItem, ItemsNumber, Move, DestPath, OpMode):
        # super().GetFiles(PanelItem, ItemsNumber, Move, DestPath, OpMode)
        if ItemsNumber == 0 or Move:
            return False
        if self.device is None:
            self.Message(["GetFiles allowed only inside device."])
            return True
        # TODO dialog with copy parameters
        # TODO progress dialog
        # TODO copy in background
        items = self.ffi.cast("struct PluginPanelItem *", PanelItem)
        DestPath = self.ffi.cast("wchar_t **", DestPath)
        dpath = self.ffi.string(DestPath[0])
        # log.debug('GetFiles: {} {} OpMode={}'.format(ItemsNumber, OpMode, dpath))
        for i in range(ItemsNumber):
            name = self.f2s(items[i].FindData.lpwszFileName)
            if name in (".", ".."):
                continue
            sqname = os.path.join(self.devicepath, name)
            dqname = os.path.join(dpath, name)
            # log.debug('pull: {} -> {} OpMode={}'.format(sqname, dqname, OpMode))
            try:
                self.devGetFile(sqname, dqname)
            except Exception as ex:
                log.exception("unknown exception:")
                msg = str(ex).split("\n")
                msg.insert(0, "Unknown exception.")
                self.Message(msg)
                return False
        return True

    def PutFiles(self, PanelItem, ItemsNumber, Move, SrcPath, OpMode):
        # super().PutFiles(PanelItem, ItemsNumber, Move, SrcPath, OpMode)
        if ItemsNumber == 0 or Move:
            return False
        if self.device is None:
            self.Message(["PetFiles allowed only inside device."])
            return True
        items = self.ffi.cast("struct PluginPanelItem *", PanelItem)
        spath = self.f2s(SrcPath)
        for i in range(ItemsNumber):
            name = self.f2s(items[i].FindData.lpwszFileName)
            if name in (".", ".."):
                continue
            sqname = os.path.join(spath, name)
            dqname = os.path.join(self.devicepath, name)
            # log.debug('push: {} -> {} OpMode={}'.format(sqname, dqname, OpMode))
            try:
                self.devPutFile(sqname, dqname)
            except Exception as ex:
                log.exception("unknown exception:")
                msg = str(ex).split("\n")
                msg.insert(0, "Unknown exception.")
                self.Message(msg)
                return False
        return True

    def DeleteFiles(self, PanelItem, ItemsNumber, OpMode):
        # super().DeleteFiles(PanelItem, ItemsNumber, OpMode)
        if ItemsNumber == 0:
            return False
        if self.device is None:
            self.Message(["DeleteFiles allowed only inside device."])
            return True
        items = self.ffi.cast("struct PluginPanelItem *", PanelItem)
        if ItemsNumber > 1:
            rc = self.ConfirmDialog("Delete", ["Do you wish to delete the files"])
        else:
            name = self.f2s(items[0].FindData.lpwszFileName)
            rc = self.ConfirmDialog("Delete", ["Do you wish to delete the file", name])
        if rc != 1:
            return True
        for i in range(ItemsNumber):
            name = self.f2s(items[i].FindData.lpwszFileName)
            if name in (".", ".."):
                continue
            dqname = os.path.join(self.devicepath, name)
            log.debug("remove: {}, OpMode={}".format(dqname, OpMode))
            try:
                self.devDelete(dqname)
            except Exception as ex:
                log.exception("unknown exception:")
                msg = str(ex).split("\n")
                msg.insert(0, "Unknown exception.")
                self.Message(msg)
                return False
        return True

    def MakeDirectory(self, Name, OpMode):
        # super().DMakeDirectory(Name, OpMode)
        if self.device is None:
            self.Message(["Make directory allowed only inside device."])
            return True
        name = self.ffi.cast("wchar_t **", Name)
        name = self.ffi.string(name[0])

        @self.ffi.callback("FARWINDOWPROC")
        def DialogProc(hDlg, Msg, Param1, Param2):
            if Msg == self.ffic.DN_INITDIALOG:
                log.debug("INITDIALOG")
                try:
                    dlg.SetText(dlg.ID_path, name)
                    dlg.SetFocus(dlg.ID_path)
                except:
                    log.exception("bang")
                log.debug("/INITDIALOG")
            elif Msg == self.ffic.DN_BTNCLICK:
                pass
            elif Msg == self.ffic.DN_KEY:
                if Param2 == self.ffic.KEY_ESC:
                    return 0
                if Param2 == self.ffic.KEY_LEFT:
                    pass
                elif Param2 == self.ffic.KEY_UP:
                    pass
                elif Param2 == self.ffic.KEY_RIGHT:
                    pass
                elif Param2 == self.ffic.KEY_DOWN:
                    pass
                elif Param2 == self.ffic.KEY_ENTER:
                    pass
                elif Param2 == self.ffic.KEY_ESC:
                    pass
            elif Msg == self.ffic.DN_MOUSECLICK:
                pass
            return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)

        b = DialogBuilder(
            self,
            DialogProc,
            "Make folder",
            "makefolder",
            0,
            VSizer(
                TEXT(None, "Create the &folder"),
                EDIT("path", 36, maxlength=40),
                HLine(),
                HSizer(
                    BUTTON("OK", "OK", default=True, flags=self.ffic.DIF_CENTERGROUP),
                    BUTTON("CANCEL", "Cancel", flags=self.ffic.DIF_CENTERGROUP),
                ),
            ),
        )
        dlg = b.build(-1, -1)

        res = self.info.DialogRun(dlg.hDlg)
        if res == dlg.ID_OK:
            path = dlg.GetText(dlg.ID_path).strip()
            dqname = os.path.normpath(os.path.join(self.devicepath, path))
            log.debug("mkdir {}".format(dqname))
            if path:
                try:
                    self.devMakeDirectory(dqname)
                except Exception as ex:
                    log.exception("unknown exception:")
                    msg = str(ex).split("\n")
                    msg.insert(0, "Unknown exception.")
                    self.Message(msg)
                    return False
        self.info.DialogFree(dlg.hDlg)

        return True

    def Execute(self):
        pass

    def ProcessKey(self, Key, ControlState):
        # log.debug("ProcessKey({0}, {1})".format(Key, ControlState))
        if ControlState == self.ffic.PKF_CONTROL:
            log.debug("ProcessKey(0x{:x}, {})".format(Key, ControlState))
        # if (
        #    Key == 0x80051 # FARMACRO_KEY_EVENT = 0x80000|1...
        #    and ControlState == self.ffic.PKF_CONTROL
        # ):
        #    # 0x80051 = CTRL+Q = quick view, here its equal to core dump when scanning /proc directory
        #    return True
        if (
            False
            and Key == self.ffic.KEY_CTRLA - self.ffic.KEY_CTRL
            and ControlState == self.ffic.PKF_CONTROL
        ):
            log.debug("ProcessKey: CTRL+A")
            # self.EditAttributes()
            return True
        if (
            Key == self.ffic.KEY_ENTER
            and ControlState == self.ffic.PKF_CONTROL
            and self.config.rexec
        ):
            log.debug("ProcessKey: CTRL+ENTER")
            self.Execute()
            return True
        return False



import logging

log = logging.getLogger(__name__)


class PluginBase:
    # private
    name = ""
    number = 0
    # public
    USERHOME = ""  # ~/.config/far2l/plugins/python
    label = ""
    # openFrom = ["DISKMENU", "PLUGINSMENU", "FINDLIST", "SHORTCUT", "COMMANDLINE", "EDITOR", "VIEWER", "FILEPANEL"]
    openFrom = []
    Configure = None  # override with method when configuration dialog is needed

    def __init__(self, parent, info, ffi, ffic):
        self.parent = parent
        self.info = info
        self.ffi = ffi
        self.ffic = ffic
        self.hplugin = self.ffi.cast("void *", id(self))

    def s2f(self, s):
        return self.ffi.new("wchar_t []", s)

    def f2s(self, s):
        return self.ffi.string(self.ffi.cast("wchar_t *", s))

    @staticmethod
    def HandleCommandLine(line):
        log.debug("Plugin.HandleCommandLine({})".format(line))
        return False

    def OpenPlugin(self, OpenFrom):
        log.debug("Plugin.OpenPlugin({}) #{}".format(OpenFrom, self.label))

    def Close(self):
        log.debug("Plugin.Close() #{}".format(self.label))

    def GetOpenPluginInfo(self, OpenInfo):
        log.debug("Plugin.GetOpenPluginInfo({}) #{}".format(OpenInfo, self.label))

    def ProcessKey(self, Key, ControlState):
        # log.debug("VFS.ProcessKey({0}, {1})".format(Key, ControlState))
        return 0


class PluginVFS(PluginBase):
    def GetFindData(self, PanelItem, ItemsNumber, OpMode):
        log.debug(
            "VFS.GetFindData({0}, {1}, {2})".format(PanelItem, ItemsNumber, OpMode)
        )
        return 0

    def FreeFindData(self, PanelItem, ItemsNumber):
        log.debug("VFS.FreeFindData({0}, {1})".format(PanelItem, ItemsNumber))

    def GetVirtualFindData(self, PanelItem, ItemsNumber, Path):
        log.debug(
            "VFS.GetVirtualFindData({0}, {1}, {2})".format(PanelItem, ItemsNumber, Path)
        )
        return 0

    def FreeVirtualFindData(self, PanelItem, ItemsNumber):
        log.debug("VFS.FreeVirtualFindData({0}, {1})".format(PanelItem, ItemsNumber))

    def Compare(self, PanelItem1, PanelItem2, Mode):
        log.debug("VFS.Compare({0}, {1}, {2})".format(PanelItem1, PanelItem2, Mode))
        return -2

    def GetFiles(self, PanelItem, ItemsNumber, Move, DestPath, OpMode):
        log.debug(
            "VFS.GetFiles({0}, {1}, {2}, {3}, {4})".format(
                PanelItem,
                ItemsNumber,
                Move,
                DestPath,
                OpMode,
            )
        )
        return 0

    def PutFiles(self, PanelItem, ItemsNumber, Move, SrcPath, OpMode):
        log.debug(
            "VFS.PutFiles({0}, {1}, {2}, {3}, {4})".format(
                PanelItem,
                ItemsNumber,
                Move,
                SrcPath,
                OpMode,
            )
        )
        return 0

    def MakeDirectory(self, Name, OpMode):
        log.debug("VFS.MakeDirectory({0}, {1})".format(Name, OpMode))
        return 0

    def SetDirectory(self, Dir, OpMode):
        log.debug("VFS.SetDirectory({0}, {1})".format(Dir, OpMode))
        return 0

    def DeleteFiles(self, PanelItem, ItemsNumber, OpMode):
        log.debug(
            "VFS.DeleteFiles({0}, {1}, {2})".format(PanelItem, ItemsNumber, OpMode)
        )
        return 0

    def SetFindList(self, PanelItem, ItemsNumber):
        log.debug("VFS.SetFindList({0}, {1})".format(PanelItem, ItemsNumber))
        return 0

    def ProcessEvent(self, Event, Param):
        log.debug("VFS.ProcessEvent({0}, {1})".format(Event, Param))
        return 0

    def ProcessHostFile(self, PanelItem, ItemsNumber, OpMode):
        log.debug(
            "VFS.ProcessHostFile({0}, {1}, {2})".format(PanelItem, ItemsNumber, OpMode)
        )
        return 0



import logging
import debugpy
from far2l.plugin import PluginBase


log = logging.getLogger(__name__)


class Config:
    configured = False
    logto = "/tmp"
    host = "127.0.0.1"
    port = 5678


class Plugin(PluginBase):
    label = "Python udebug"
    openFrom = ["PLUGINSMENU", "COMMANDLINE", "EDITOR", "VIEWER", "FILEPANEL"]

    def debug(self):
        if Config.configured:
            log.debug("udebug can be configured only once")
            return

        if not Config.configured:
            Config.configured = True
            debugpy.log_to(Config.logto)
            # in vs code debuger select attach, port = 5678
            # commands in shell:
            #   py:debug
            # elsewhere in python code:
            #   import debugpy
            #   debugpy.breakpoint()
            debugpy.listen((Config.host, Config.port))
        debugpy.wait_for_client()
        # debugpy.breakpoint()

    def breakpoint(self):
        debugpy.breakpoint()

    @staticmethod
    def HandleCommandLine(line):
        return line in ("debug", "breakpoint")

    def CommandLine(self, line):
        getattr(self, line)()

    def Configure(self):
        @self.ffi.callback("FARWINDOWPROC")
        def KeyDialogProc(hDlg, Msg, Param1, Param2):
            # if Msg == self.ffic.DN_KEY:
            #    if Param2 == 27:
            #        return 0
            return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)

        fdis = [
            (
                self.ffic.DI_DOUBLEBOX,
                2,
                1,
                32,
                7,
                0,
                {"Selected": 0},
                0,
                0,
                self.s2f("Python Debug Config"),
                0,
            ),
            (
                self.ffic.DI_TEXT,
                4,
                2,
                11,
                0,
                0,
                {"Selected": 0},
                0,
                0,
                self.s2f("Log path"),
                0,
            ),
            (
                self.ffic.DI_EDIT,
                13,
                2,
                30,
                0,
                0,
                {"Selected": 0},
                0,
                0,
                self.s2f(Config.logto),
                32,
            ),
            (
                self.ffic.DI_TEXT,
                4,
                3,
                8,
                0,
                0,
                {"Selected": 0},
                0,
                0,
                self.s2f("Host"),
                0,
            ),
            (
                self.ffic.DI_EDIT,
                13,
                3,
                30,
                0,
                0,
                {"Selected": 0},
                0,
                0,
                self.s2f(Config.host),
                32,
            ),
            (
                self.ffic.DI_TEXT,
                4,
                4,
                8,
                0,
                0,
                {"Selected": 0},
                0,
                0,
                self.s2f("Port"),
                0,
            ),
            (
                self.ffic.DI_EDIT,
                13,
                4,
                30,
                0,
                0,
                {"Selected": 0},
                0,
                0,
                self.s2f(str(Config.port)),
                5,
            ),
            (
                self.ffic.DI_TEXT,
                2,
                5,
                0,
                0,
                0,
                {"Selected": 0},
                self.ffic.DIF_SEPARATOR,
                0,
                self.ffi.NULL,
                0,
            ),
            (
                self.ffic.DI_BUTTON,
                6,
                6,
                0,
                0,
                0,
                {"Selected": 0},
                0,
                self.ffic.DIF_DEFAULT,
                self.s2f("OK"),
                0,
            ),
            (
                self.ffic.DI_BUTTON,
                19,
                6,
                0,
                0,
                0,
                {"Selected": 0},
                0,
                0,
                self.s2f("Cancel"),
                0,
            ),
        ]
        fdi = self.ffi.new("struct FarDialogItem []", fdis)
        hDlg = self.info.DialogInit(
            self.info.ModuleNumber,
            -1,
            -1,
            fdis[0][3] + 3,
            fdis[0][4] + 2,
            self.s2f("Debug Config"),
            fdi,
            len(fdi),
            0,
            0,
            KeyDialogProc,
            0,
        )
        res = self.info.DialogRun(hDlg)
        log.debug("dialog result: {0}".format(res))
        if res == 8:
            result = []
            for i in (2, 4, 6):
                sptr = self.info.SendDlgMessage(
                    hDlg, self.ffic.DM_GETCONSTTEXTPTR, i, 0
                )
                result.append(self.f2s(sptr))
            log.debug("result: {0}".format(result))
            Config.logto = result[0]
            Config.host = result[1]
            Config.port = int(result[2])
        self.info.DialogFree(hDlg)
        log.debug("end dialog")

    def OpenPlugin(self, OpenFrom):
        self.Configure()
